//  Boost endian.hpp header file (proposed) ----------------------------------//

//  (C) Copyright Scott McMurray 2006

//  Use, modification, and distribution is subject to the Boost Software
//  License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy
//  at http://www.boost.org/LICENSE_1_0.txt)

//  See library home page at http://www.boost.org/libs/???????

//----------------------------------------------------------------------------//

#ifndef BOOST_ENDIAN_HPP
#define BOOST_ENDIAN_HPP

#include <boost/detail/endian.hpp>
//#include <boost/cast.hpp>
#include <boost/type_traits/is_signed.hpp>
#include <boost/static_assert.hpp>
#include <boost/integer.hpp>
#include <limits>
#include <cstring>
//#include <boost/integer_cover_operators.hpp>
#include "integer_cover_operators.hpp"
#include <algorithm>
#include "exact.hpp"

namespace boost {

template <typename T>
struct byte_reverser {
    BOOST_STATIC_ASSERT( std::numeric_limits<T>::is_specialized );
    BOOST_STATIC_ASSERT( std::numeric_limits<T>::is_integer );
    BOOST_STATIC_ASSERT( std::numeric_limits<T>::is_bounded );
    void operator()(T &t) {
        std::reverse( reinterpret_cast<char*>(&t),
                      reinterpret_cast<char*>(&t)
                     + detail::round_up_div< std::numeric_limits<T>::digits,
                                             detail::bits_per_byte >::value );
    }
};

template <typename T>
void byte_reverse (T &t) {
    byte_reverser<T> r;
    r(t);
}

template <typename T>
class native_endian
 : boost::integer_cover_operators< native_endian<T>,
                                   typename boost::wrapped_type<T>::type > {
  public:
    typedef T value_type;
    typedef typename boost::wrapped_type<T>::type wrapped_type;
    value_type &get() { return value; }
    value_type const &get() const { return value; }
    native_endian() {}
    native_endian(wrapped_type i) : value(i) {}
    operator wrapped_type const &() const { return value; }
  private:
    value_type value;
};

template <typename T>
class little_endian
 : boost::integer_cover_operators< little_endian<T>,
                                   typename boost::wrapped_type<T>::type > {
  public:
    typedef T value_type;
    typedef typename boost::wrapped_type<T>::type wrapped_type;
    value_type &get() { return value; }
    value_type const &get() const { return value; }
    little_endian() {}
  #ifdef BOOST_BIG_ENDIAN
    little_endian(wrapped_type i) {
        value = i;
        byte_reverse(value);
    }
    operator wrapped_type() const {
        value_type i( value );
        byte_reverse(i);
        return i;
    }
  #else
    little_endian(wrapped_type i) : value(i) {}
    operator wrapped_type const &() const { return value; }
  #endif
  private:
    value_type value;
};

template <typename T>
class big_endian 
 : boost::integer_cover_operators< big_endian<T>,
                                   typename boost::wrapped_type<T>::type > {
  public:
    typedef T value_type;
    typedef typename boost::wrapped_type<T>::type wrapped_type;
    value_type &get() { return value; }
    value_type const &get() const { return value; }
    big_endian() {}
  #ifdef BOOST_LITTLE_ENDIAN
    big_endian(wrapped_type i) {
        value = i;
        byte_reverse(value);
    }
    operator wrapped_type() const {
        value_type i( value );
        byte_reverse(i);
        return i;
    }
  #else
    big_endian(wrapped_type i) : value(i) {}
    operator wrapped_type() const { return value; }
  #endif
  private:
    value_type value;
};

} // namespace boost

#endif // BOOST_ENDIAN_HPP
